//  *****************************************************************************
//   						main.c
// 
//              Demonstracny program pre dosku KTK ARM7
//
//              Meranie vzdialenosti ultrazvukom, zobrazovanie na seriovom terminaly
//
//  *****************************************************************************

// Hlavickove subory
#include "board.h"

#define STOP    1 << 1                  // vystup z komparatora - PA1
#define USART1_BAUD_RATE 115200         // rychlost USART1
#define STEPS   ((unsigned int) ((MCK/8) + 500000) / 1000000L)    // pocet krokov casovaca na 1 us

// Priorita preruseni
#define   PWMC_INTERRUPT_LEVEL  0       // najnizsia priorita prerusenia
#define   PIO_INTERRUPT_LEVEL   1       // vyssia priorita prerusenia

// Globalne premenne
int pocet_period=0;           // pocitanie vyslanych period
int f_stop;                   // priznak prijatia odrazeneho signalu

//  *****************************************************
//  Mikrosekundove oneskorenie (max. cca 10 000)
//  *****************************************************
void delay(unsigned int us)
{
  AT91C_BASE_PMC->PMC_PCER = (1 << AT91C_ID_TC0);  // povolenie hodin pre TC0
  AT91C_BASE_TC0->TC_CMR = 1;                      // volba hodin (MCK/8)
  AT91C_BASE_TC0->TC_RC = us * STEPS;              // nastavenie porovnavacieho registra C
  AT91C_BASE_TC0->TC_CCR = AT91C_TC_CLKEN |        // povolenie hodin
                           AT91C_TC_SWTRG;         // restart casovaca

  // cakanie na ukoncenie oneskorenia
  // ked je nastaveny compare C flag, je hotovo
  while ((AT91C_BASE_TC0->TC_SR & AT91C_TC_CPCS) == 0) asm ("nop");
    
  AT91C_BASE_PMC->PMC_PCDR = (1 << AT91C_ID_TC0);  // zakazanie hodin pre TC0
}

//  *******************************************************
//  Obsluha prerusenia od PWM0
//  *******************************************************
void PWM0_prerusenie(void)
{
  // zvysenie poctu period
  pocet_period++;
  // kontrola konca
  if (pocet_period > 20)
  { pocet_period = 0;
    AT91C_BASE_PWMC->PWMC_DIS = 1;    // zakazanie kanala PWM0
  }

  // precitanie stavu preruseni kvoli povoleniu opetovneho prerusenia
  AT91C_BASE_PWMC->PWMC_ISR;
}

//  *******************************************************
//  Obsluha prerusenia od PIO
//  *******************************************************
void PIO_prerusenie(void)
{
  // zastavenie casovaca TC1
  AT91C_BASE_TC1->TC_CCR = AT91C_TC_CLKDIS;

  f_stop = 1;                            // priznak prijatia odrazeneho signalu 

  // precitanie stavu preruseni kvoli povoleniu opetovneho prerusenia
  AT91C_BASE_PIOA->PIO_ISR;
}

//  *******************************************************
//  Vyslanie znaku cez USART1
//  *******************************************************
void USART1_posli_znak(char znak)
{
  // cakanie na vyslanie predosleho znaku
  while(AT91C_BASE_US1->US_CSR & AT91C_US_TXRDY == 0);
  // vyslanie znaku
  AT91C_BASE_US1->US_THR = znak;
  // chvilka cakania
  delay(1000);
}

//  *******************************************************
//  Inicializacia USART1
//  *******************************************************
void USART1_init(void)
{
  // Konfiguracia PIO radica do rezimu periferie SPI
  // PIO ma zakazany pristup na dane piny
  AT91C_BASE_PIOA->PIO_PDR = AT91C_PA21_RXD1 | AT91C_PA22_TXD1;
  // povolenie periferie A
  AT91C_BASE_PIOA->PIO_ASR = AT91C_PA21_RXD1 | AT91C_PA22_TXD1;

  // povolenie hodin pre USART
  AT91C_BASE_PMC->PMC_PCER = 1 << AT91C_ID_US1;

  // konfiguracia USART
  AT91C_BASE_US1->US_IDR = 0xffff;        // zakazanie preruseni
  AT91C_BASE_US1->US_CR = AT91C_US_RSTRX | AT91C_US_RSTTX | AT91C_US_RXDIS | AT91C_US_TXDIS;

  AT91C_BASE_US1->US_BRGR = MCK/(USART1_BAUD_RATE * 16); // nastavenie rychlosti
  AT91C_BASE_US1->US_MR = AT91C_US_CHRL_8_BITS |    // 8 bitove udaje
                          AT91C_US_NBSTOP_1_BIT |   // 1 stop bit
                          AT91C_US_PAR_NONE;        // ziadna parita
  // povolenie vysielania znakov
  AT91C_BASE_US1->US_CR = AT91C_US_TXEN;
}

//  *******************************************************
//                Hlavny program
//  *******************************************************
int main(void) 
{ int x, vzd;

  //******************* Uvodna inicializacia *******************
  // Povolenie Resetu - minimalna dlzka 960 us
  AT91C_BASE_RSTC->RSTC_RMR = AT91C_RSTC_URSTEN | (0x4<<8) | (unsigned int)(0xA5<<24);
	
  //******************* Inicializacia portu A ******************
  // povolenie hodin pre PIOA
  AT91C_BASE_PMC->PMC_PCER = (unsigned int) 1 << AT91C_ID_PIOA; 
	
  // konfiguracia jednotlivych pinov
  // vstupne piny
  AT91C_BASE_PIOA->PIO_ODR = S2 | STOP;
  // vystupne piny
  AT91C_BASE_PIOA->PIO_OER = D1;
  // povolenie Pull-Up rezistoru
  AT91C_BASE_PIOA->PIO_PPUER = S2;
  // povolenie vsetkych nakonfigurovanych pinov
  AT91C_BASE_PIOA->PIO_PER = D1 | S2 | STOP;

  //******************* Nastavenie prerusenia od PIO ******************
  // nastavenie obsluzneho programu
  AT91C_BASE_AIC->AIC_SVR[AT91C_ID_PIOA] = (unsigned int) PIO_prerusenie;
  // zdroj prerusenia a priorita
  AT91C_BASE_AIC->AIC_SMR[AT91C_ID_PIOA] = PIO_INTERRUPT_LEVEL;
  // zrusenie prerusenia (ak nejake nahodou bolo)
  AT91C_BASE_AIC->AIC_ICCR = 0x1 << AT91C_ID_PIOA;
  // povolenie prerusenia pre prislusny pin PIOA
  AT91C_BASE_PIOA->PIO_IER = STOP;       // pin PA1
  // povolenie prerusenia od PIO
  AT91C_BASE_AIC->AIC_IECR = 0x1 << AT91C_ID_PIOA;

  //******************* Nastavenie PWM ******************
  // Konfiguracia PIO radica do rezimu periferie PWM (kanal PWM0)
  // PIO ma zakazany pristup na dany pin
  AT91C_BASE_PIOA->PIO_PDR = AT91C_PA0_PWM0;
  // povolenie periferie A
  AT91C_BASE_PIOA->PIO_ASR = AT91C_PA0_PWM0;
  // povolenie hodin pre PWM
  AT91C_BASE_PMC->PMC_PCER = (unsigned int) 1 << AT91C_ID_PWMC;
  // zakazanie vsetkych kanalov
  AT91C_BASE_PWMC->PWMC_DIS = 0x0f;
  // nastavenie modu pre PWM0 - zarovnanie signalu vlavo
  AT91C_BASE_PWMC->PWMC_CH[0].PWMC_CMR = AT91C_PWMC_CPOL;  // volba hodin - MCK, polarita = 1 (signal zacina v 1)
  // nastavenie periody: X/MCK = 1198/47923200 = 25 us (40 kHz)
  AT91C_BASE_PWMC->PWMC_CH[0].PWMC_CPRDR = 1198;
  // nastavenie koeficientu plnenia presne na polovicu (Duty Cycle)
  AT91C_BASE_PWMC->PWMC_CH[0].PWMC_CDTYR = 599;
  // povolenie prerusenia od PWM0
  AT91C_BASE_PWMC->PWMC_IER = 1;

  // nastavenie obsluzneho programu
  AT91C_BASE_AIC->AIC_SVR[AT91C_ID_PWMC] = (unsigned int) PWM0_prerusenie;
  // zdroj prerusenia a priorita
  AT91C_BASE_AIC->AIC_SMR[AT91C_ID_PWMC] = PWMC_INTERRUPT_LEVEL;
  // zrusenie prerusenia (ak nejake nahodou bolo)
  AT91C_BASE_AIC->AIC_ICCR = 0x1 << AT91C_ID_PWMC;
  // povolenie prerusenia od PWMC
  AT91C_BASE_AIC->AIC_IECR = 0x1 << AT91C_ID_PWMC;

  // Casovac TC1 sluzi na pocitanie casu odrazeneho signalu    
  AT91C_BASE_PMC->PMC_PCER = (1 << AT91C_ID_TC1);  // povolenie hodin pre TC1
  AT91C_BASE_TC1->TC_CMR = 1;                      // volba hodin (MCK/8)
  AT91C_BASE_TC1->TC_RC = 0xffff;                  // nastavenie porovnavacieho registra C na maximum 
                                                   // kvoli kontrole timeoutu
  //******************* Hlavna cast programu *******************
  // LED nesvieti
  AT91C_BASE_PIOA->PIO_CODR = D1;

  // inicializacia USART 
  USART1_init();

  // nekonecna slucka
  while(1)
  { AT91C_BASE_PIOA->PIO_CODR = D1;	      // zhasnutie LED
    // cakanie na stlacenie S2  
    while(AT91C_BASE_PIOA->PIO_PDSR & S2)     // kontrola, ci je S2 v log.1
      delay(200);                             // cakanie kvoli potlaceniu zakmitov
    // cakanie na pustenie S2
    while(!(AT91C_BASE_PIOA->PIO_PDSR & S2))  // kontrola, ci je S2 v log.0
      delay(200);                             // cakanie kvoli potlaceniu zakmitov

    f_stop = 0;
    // povolenie casovaca TC1 a jeho spustenie
    AT91C_BASE_TC1->TC_CCR = AT91C_TC_CLKEN | AT91C_TC_SWTRG;

    // start PWM0 (vysielanie 20 period 40 kHz signalu)
    AT91C_BASE_PWMC->PWMC_ENA = 1;

    // cakanie na odrazeny signal alebo na timeout (ak je nastaveny compare C flag)
    while (((AT91C_BASE_TC1->TC_SR & AT91C_TC_CPCS) == 0) && (f_stop == 0));

    if (f_stop)
    { AT91C_BASE_PIOA->PIO_SODR = D1;	      // zapnutie LED
      // nacitanie casu
      x = AT91C_BASE_TC1->TC_CV;
      // vypocet vzdialenosti: t=x/2*8/MCK (8/MCK je perioda casovaca TC1, x/2 - pocitame len polovicu casu)
      // s=v*t (v = 340 m/s = 340 000 mm/s)
      // konstanta 50 bola zistena kalibraciou - oneskorenie detekcie signalu
      vzd = x/2*340000*8/MCK - 50;
      // vyslanie cez seriovy port
      USART1_posli_znak(vzd/100 + '0');         // stovky
      USART1_posli_znak((vzd%100)/10 + '0');    // desiatky
      USART1_posli_znak(((vzd%100)%10) + '0');  // jednotky
      USART1_posli_znak('m');
      USART1_posli_znak('m');
      USART1_posli_znak(13);                    // CR
      USART1_posli_znak(10);                    // LF - novy riadok
    }
    else
      AT91C_BASE_TC1->TC_CCR = AT91C_TC_CLKDIS;   // zastavenie hodin pre casovac
  }
}
